#define CL_HPP_TARGET_OPENCL_VERSION 300

#include <vector>
#include <iostream>
#include <fstream>
#include <cmath>
#include <CL/opencl.hpp>

#include "../revengmc.hpp"

#define MAX_48 0xFFFFFFFFFFFF
#define WORKER_COUNT 1024
#define MAGIC_NUM 25214903917

// const size_t GROUP_SIZE = 1 << 10ul;
// const size_t BATCH_SIZE = 1ul << 40ul;

const size_t GROUP_SIZE = 1 << 10;
const size_t BATCH_SIZE = 1ull << 36;

static cl_int err;

static void handle(int err, int skip = 0) {
	if (err != CL_SUCCESS) {
		std::string strerr;

		switch (err) {
			case CL_DEVICE_NOT_FOUND: strerr = "CL_DEVICE_NOT_FOUND"; break;
			case CL_DEVICE_NOT_AVAILABLE: strerr = "CL_DEVICE_NOT_AVAILABLE"; break;
			case CL_COMPILER_NOT_AVAILABLE: strerr = "CL_COMPILER_NOT_AVAILABLE"; break;
			case CL_MEM_OBJECT_ALLOCATION_FAILURE: strerr = "CL_MEM_OBJECT_ALLOCATION_FAILURE"; break;
			case CL_OUT_OF_RESOURCES: strerr = "CL_OUT_OF_RESOURCES"; break;
			case CL_OUT_OF_HOST_MEMORY: strerr = "CL_OUT_OF_HOST_MEMORY"; break;
			case CL_PROFILING_INFO_NOT_AVAILABLE: strerr = "CL_PROFILING_INFO_NOT_AVAILABLE"; break;
			case CL_MEM_COPY_OVERLAP: strerr = "CL_MEM_COPY_OVERLAP"; break;
			case CL_IMAGE_FORMAT_MISMATCH: strerr = "CL_IMAGE_FORMAT_MISMATCH"; break;
			case CL_IMAGE_FORMAT_NOT_SUPPORTED: strerr = "CL_IMAGE_FORMAT_NOT_SUPPORTED"; break;
			case CL_BUILD_PROGRAM_FAILURE: strerr = "CL_BUILD_PROGRAM_FAILURE"; break;
			case CL_MAP_FAILURE: strerr = "CL_MAP_FAILURE"; break;
			case CL_MISALIGNED_SUB_BUFFER_OFFSET: strerr = "CL_MISALIGNED_SUB_BUFFER_OFFSET"; break;
			case CL_COMPILE_PROGRAM_FAILURE: strerr = "CL_COMPILE_PROGRAM_FAILURE"; break;
			case CL_LINKER_NOT_AVAILABLE: strerr = "CL_LINKER_NOT_AVAILABLE"; break;
			case CL_LINK_PROGRAM_FAILURE: strerr = "CL_LINK_PROGRAM_FAILURE"; break;
			case CL_DEVICE_PARTITION_FAILED: strerr = "CL_DEVICE_PARTITION_FAILED"; break;
			case CL_KERNEL_ARG_INFO_NOT_AVAILABLE: strerr = "CL_KERNEL_ARG_INFO_NOT_AVAILABLE"; break;
			case CL_INVALID_VALUE: strerr = "CL_INVALID_VALUE"; break;
			case CL_INVALID_DEVICE_TYPE: strerr = "CL_INVALID_DEVICE_TYPE"; break;
			case CL_INVALID_PLATFORM: strerr = "CL_INVALID_PLATFORM"; break;
			case CL_INVALID_DEVICE: strerr = "CL_INVALID_DEVICE"; break;
			case CL_INVALID_CONTEXT: strerr = "CL_INVALID_CONTEXT"; break;
			case CL_INVALID_QUEUE_PROPERTIES: strerr = "CL_INVALID_QUEUE_PROPERTIES"; break;
			case CL_INVALID_COMMAND_QUEUE: strerr = "CL_INVALID_COMMAND_QUEUE"; break;
			case CL_INVALID_HOST_PTR: strerr = "CL_INVALID_HOST_PTR"; break;
			case CL_INVALID_MEM_OBJECT: strerr = "CL_INVALID_MEM_OBJECT"; break;
			case CL_INVALID_IMAGE_FORMAT_DESCRIPTOR: strerr = "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR"; break;
			case CL_INVALID_IMAGE_SIZE: strerr = "CL_INVALID_IMAGE_SIZE"; break;
			case CL_INVALID_SAMPLER: strerr = "CL_INVALID_SAMPLER"; break;
			case CL_INVALID_BINARY: strerr = "CL_INVALID_BINARY"; break;
			case CL_INVALID_BUILD_OPTIONS: strerr = "CL_INVALID_BUILD_OPTIONS"; break;
			case CL_INVALID_PROGRAM: strerr = "CL_INVALID_PROGRAM"; break;
			case CL_INVALID_PROGRAM_EXECUTABLE: strerr = "CL_INVALID_PROGRAM_EXECUTABLE"; break;
			case CL_INVALID_KERNEL_NAME: strerr = "CL_INVALID_KERNEL_NAME"; break;
			case CL_INVALID_KERNEL_DEFINITION: strerr = "CL_INVALID_KERNEL_DEFINITION"; break;
			case CL_INVALID_KERNEL: strerr = "CL_INVALID_KERNEL"; break;
			case CL_INVALID_ARG_INDEX: strerr = "CL_INVALID_ARG_INDEX"; break;
			case CL_INVALID_ARG_VALUE: strerr = "CL_INVALID_ARG_VALUE"; break;
			case CL_INVALID_ARG_SIZE: strerr = "CL_INVALID_ARG_SIZE"; break;
			case CL_INVALID_KERNEL_ARGS: strerr = "CL_INVALID_KERNEL_ARGS"; break;
			case CL_INVALID_WORK_DIMENSION: strerr = "CL_INVALID_WORK_DIMENSION"; break;
			case CL_INVALID_WORK_GROUP_SIZE: strerr = "CL_INVALID_WORK_GROUP_SIZE"; break;
			case CL_INVALID_WORK_ITEM_SIZE: strerr = "CL_INVALID_WORK_ITEM_SIZE"; break;
			case CL_INVALID_GLOBAL_OFFSET: strerr = "CL_INVALID_GLOBAL_OFFSET"; break;
			case CL_INVALID_EVENT_WAIT_LIST: strerr = "CL_INVALID_EVENT_WAIT_LIST"; break;
			case CL_INVALID_EVENT: strerr = "CL_INVALID_EVENT"; break;
			case CL_INVALID_OPERATION: strerr = "CL_INVALID_OPERATION"; break;
			case CL_INVALID_GL_OBJECT: strerr = "CL_INVALID_GL_OBJECT"; break;
			case CL_INVALID_BUFFER_SIZE: strerr = "CL_INVALID_BUFFER_SIZE"; break;
			case CL_INVALID_MIP_LEVEL: strerr = "CL_INVALID_MIP_LEVEL"; break;
			case CL_INVALID_GLOBAL_WORK_SIZE: strerr = "CL_INVALID_GLOBAL_WORK_SIZE"; break;
			case CL_INVALID_PROPERTY: strerr = "CL_INVALID_PROPERTY"; break;
			case CL_INVALID_IMAGE_DESCRIPTOR: strerr = "CL_INVALID_IMAGE_DESCRIPTOR"; break;
			case CL_INVALID_COMPILER_OPTIONS: strerr = "CL_INVALID_COMPILER_OPTIONS"; break;
			case CL_INVALID_LINKER_OPTIONS: strerr = "CL_INVALID_LINKER_OPTIONS"; break;
			case CL_INVALID_DEVICE_PARTITION_COUNT: strerr = "CL_INVALID_DEVICE_PARTITION_COUNT"; break;
			default: strerr = "UNKNOWN";
		}

		std::cerr << "Error " << strerr << "(" << err << ")!\n";
		std::exit(err);
	}
}
static void handle() {
	handle(err, 1);
}

static std::string read_file(const std::string &path) {
	std::ifstream f { path };
	if (f.fail()) return nullptr;

	f.seekg(0, std::ios_base::end);
	size_t n = f.tellg();
	f.seekg(0, std::ios_base::beg);

	auto buff = new char[n + 1];
	buff[n] = 0;
	f.read(buff, n);

	std::string res { buff };
	return res;
}

static cl::Context make_context() {
	// cl::Platform platform = cl::Platform::get(&err); handle();
	// cl_context_properties props[] = { CL_CONTEXT_PLATFORM, (cl_context_properties)platform(), 0 };
	// auto context = cl::Context(CL_DEVICE_TYPE_ALL, props, nullptr, nullptr, &err);
	auto context = cl::Context::getDefault(&err);
	handle();

	return context;
}

struct impl_t {
	cl::Context ctx;
	cl::Kernel kernel;
	cl::CommandQueue queue;
	cl::Buffer buffer;
	cl::Event event;

	std::vector<condition_t> conditions;

	std::vector<uint64_t> results;
};

struct args_t {
	uint64_t task_n;
	uint64_t start;
	uint64_t group_size;
};

static impl_t *impl;

void engine::init(const std::vector<condition_t> &tasks) {
	if (impl != nullptr) {
		std::cerr << "Backend already initialized";
		std::exit(1);
	}

	impl = new impl_t();
	impl->ctx = make_context();
	impl->conditions = tasks;

	auto device = impl->ctx.getInfo<CL_CONTEXT_DEVICES>()[0];

	impl->queue = cl::CommandQueue(impl->ctx, device, 0, &err); handle();

	size_t size = sizeof(condition_t) * impl->conditions.size();
	impl->buffer = cl::Buffer(impl->ctx, CL_MEM_READ_WRITE, size, nullptr, &err); handle();
	handle(impl->queue.enqueueWriteBuffer(impl->buffer, true, 0, size, &impl->conditions[0], nullptr, &impl->event));
	impl->event.wait();

	auto program = cl::Program(impl->ctx, read_file("revengmc-kernel.cl")); handle();
	auto build_err = program.build();

	if (build_err != CL_SUCCESS) {
		auto build_log = program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(&err); handle();
		std::cerr << build_log[0].second;
		std::exit(1);
	}

	impl->kernel = cl::Kernel(program, "run", &err); handle();
}

size_t engine::batch_size() {
	return BATCH_SIZE;
}

const std::vector<uint64_t> &engine::run(size_t start, size_t end) {
	size_t n = end - start;
	size_t groups = n / GROUP_SIZE;
	if (groups < GROUP_SIZE) groups = 0;
	
	cl::Event event;
	handle(impl->kernel.setArg(0, args_t {
		.task_n = impl->conditions.size(),
		.start = start,
		.group_size = GROUP_SIZE,
	}));
	handle(impl->kernel.setArg(1, impl->buffer));
	handle(impl->queue.enqueueNDRangeKernel(impl->kernel, cl::NullRange, cl::NDRange(groups * GROUP_SIZE), cl::NDRange(GROUP_SIZE), nullptr, &event));
	event.wait();
 
	if (groups * GROUP_SIZE < n) {
		cl::Event event;

		handle(impl->kernel.setArg(0, args_t {
			.task_n = impl->conditions.size(),
			.start = start + groups * GROUP_SIZE,
			.group_size = 1,
		}));
		handle(impl->kernel.setArg(1, impl->buffer));
		handle(impl->queue.enqueueNDRangeKernel(impl->kernel, cl::NullRange, cl::NDRange(n - groups * GROUP_SIZE), cl::NDRange(1), nullptr, &event));
		event.wait();
	}

	return impl->results;
}


void engine::free() {
	if (impl == nullptr) {
		std::cerr << "Backend already freed";
		std::exit(1);
	}

	delete impl;
}
