package me.topchetoeu.revengmc.core;

import java.io.File;

import net.minecraft.CrashReport;
import net.minecraft.SharedConstants;
import net.minecraft.Util;
import net.minecraft.commands.Commands.CommandSelection;
import net.minecraft.core.Registry;
import net.minecraft.core.RegistryAccess;
import net.minecraft.server.Bootstrap;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.ServerResources;
import net.minecraft.server.packs.PackType;
import net.minecraft.server.packs.repository.FolderRepositorySource;
import net.minecraft.server.packs.repository.PackRepository;
import net.minecraft.server.packs.repository.PackSource;
import net.minecraft.server.packs.repository.ServerPacksSource;
import net.minecraft.world.level.DataPackConfig;
import net.minecraft.world.level.dimension.DimensionType;

public record MinecraftContext(
	RegistryAccess.RegistryHolder registry,
	PackRepository packRepo,
	ServerResources resources,
	DimensionType overworldLocation
) {
	public static MinecraftContext initialize() throws Throwable {
		SharedConstants.tryDetectVersion();

		CrashReport.preload();
		Bootstrap.bootStrap();
		Bootstrap.validate();

		var registry = RegistryAccess.builtin();

		var packRepo = new PackRepository(
			PackType.SERVER_DATA, new ServerPacksSource(),
			new FolderRepositorySource(new File("world/datapacks"), PackSource.WORLD)
		);

		MinecraftServer.configurePackRepository(packRepo, DataPackConfig.DEFAULT, true);

		var resources = ServerResources.loadResources(
			packRepo.openAllSelected(),
			registry,
			CommandSelection.DEDICATED, 4,
			Util.backgroundExecutor(),
			Runnable::run
		).get();
		resources.updateGlobals();

		var overworldLocation = registry.registryOrThrow(Registry.DIMENSION_TYPE_REGISTRY).get(DimensionType.OVERWORLD_LOCATION);
		if (overworldLocation == null) throw new RuntimeException("wtf");

		return new MinecraftContext(registry, packRepo, resources, overworldLocation);
	}
}