package me.topchetoeu.revengmc.search;

import java.io.IOException;

import me.topchetoeu.revengmc.core.BogusServer;
import me.topchetoeu.revengmc.core.MinecraftContext;
import net.minecraft.core.SectionPos;
import net.minecraft.server.level.ServerLevel;
import net.minecraft.world.level.ChunkPos;
import net.minecraft.world.level.StructureFeatureManager;
import net.minecraft.world.level.biome.BiomeSource;
import net.minecraft.world.level.chunk.ChunkGenerator;
import net.minecraft.world.level.chunk.ProtoChunk;
import net.minecraft.world.level.chunk.UpgradeData;
import net.minecraft.world.level.levelgen.feature.ConfiguredStructureFeature;
import net.minecraft.world.level.levelgen.structure.StructureStart;
import net.minecraft.world.level.levelgen.structure.templatesystem.StructureManager;

public class Tester implements AutoCloseable {
	public final ServerLevel level;
	public final long seed;
	public final StructureFeatureManager features;
	public final StructureManager structures;
	public final ChunkGenerator chunkGen;
	public final BiomeSource biomeSrc;
	public final BogusServer server;
	public final MinecraftContext ctx;

	public boolean placementPossible(ConfiguredStructureFeature<?, ?> structure, int chunkX, int chunkY) {
		var chunkPos = new ChunkPos(chunkX, chunkY);
		var chunk = new ProtoChunk(chunkPos, UpgradeData.EMPTY, level);
		var biome = biomeSrc.getPrimaryBiome(chunk.getPos());

		var sectionPos = SectionPos.bottomOf(chunk);
		var start = features.getStartForFeature(sectionPos, structure.feature, chunk);
		int refs = start != null ? start.getReferences() : 0;
		var config = chunkGen.getSettings().getConfig(structure.feature);

		if (config != null) {
			var newStart = structure.generate(ctx.registry(), chunkGen, biomeSrc, structures, seed, chunkPos, biome, refs, config, level);
			if (newStart == StructureStart.INVALID_START) return false;
			features.setStartForFeature(sectionPos, structure.feature, newStart, chunk);
			return true;
		}

		return false;
	}

	public Tester(MinecraftContext ctx, long seed) throws IOException {
		this.server = new BogusServer(ctx, seed);
		this.ctx = ctx;

		this.level = server.level;
		this.seed = seed;

		this.chunkGen = level.getChunkSource().getGenerator();
		this.features = level.structureFeatureManager();
		this.structures = level.getStructureManager();
		this.biomeSrc = chunkGen.getBiomeSource();
	}

	@Override public void close() throws IOException {
		level.close();
		server.close();
	}
}